/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.wsdl.Definition;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.PortType;
import javax.wsdl.WSDLException;
import javax.xml.namespace.QName;

import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.tx.Transaction;
import org.jbpm.xml.Parse;
import org.ow2.orchestra.deployment.Deployer;
import org.ow2.orchestra.deployment.Deployment;
import org.ow2.orchestra.execution.BpelExecution;
import org.ow2.orchestra.execution.BpelProcess;
import org.ow2.orchestra.execution.services.MessageCarrier;
import org.ow2.orchestra.execution.services.OperationKey;
import org.ow2.orchestra.execution.services.Receiver;
import org.ow2.orchestra.execution.services.itf.Invoker;
import org.ow2.orchestra.parsing.BpelImportParser;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.WsdlUtil;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.engine.orchestra.jbi.JBIInvoker;
import org.ow2.petals.engine.orchestra.listener.JBIListener;
import org.ow2.petals.engine.orchestra.util.Marshaller;
import org.w3c.dom.Document;

/**
 * This class implements a context for BPEL
 *
 * @author Nicolas SALATGE - eBMWebSourcing
 *
 */
public final class BPELContext {

	/**
	 * BPEL deployment
	 */
	private Deployment bpelDeployment;

	/**
	 * BPEL deployer
	 */
	private Deployer bpelDeployer;

	/**
	 * BPEL process
	 */
	private BpelProcess bpelProcess;

	/**
	 * The jbi environment for orchestra
	 */
	private Environment jbiEnvironment;

	/**
	 * the environment Factory
	 */
	private final EnvironmentFactory environmentFactory;

	/**
	 * JDK logger
	 */
	private final Logger logger;

	/**
	 * BPEL process name
	 */
	private final URL bpelProcessName;

	/**
	 * WSDL name
	 */
	private final URL wsdlName;

	/**
	 * WSDL Definition
	 */
	private final Definition wsdlDef;

	/**
	 * Externals WSDLs
	 */
	private final List<URL> externalsWsdls;

	/**
	 * Default constructor
	 *
	 * @param bpelProcessName
	 *            the URL of bpel process
	 * @param wsdlName
	 *            the URL of wsdl
	 * @param environmentFactory
	 *            the jbi environnement factory
	 * @param externalsWsdls
	 *            the externals wsdl
	 * @param logger
	 *            the logger
	 */
	public BPELContext(final URL bpelProcessName, final URL wsdlName,
			final EnvironmentFactory environmentFactory, final Logger logger,
			final ArrayList<URL> externalsWsdls) {

		// Initialization
		this.environmentFactory = environmentFactory;
		this.logger = logger;
		this.bpelDeployment = null;
		this.bpelDeployer = null;

		this.bpelProcessName = bpelProcessName;
		this.wsdlName = wsdlName;
		this.wsdlDef = WsdlUtil.readWsdl(this.wsdlName);
		this.bpelProcess = null;
		this.externalsWsdls = externalsWsdls;

		this.openContext();

		// this.invocations = new ArrayList<ServiceInvocation>();
	}

	/**
	 * get the logger
	 *
	 * @return the logger
	 */
	public Logger getLogger() {
		return this.logger;
	}

	/**
	 * Deploy the bpel process
	 *
	 * @throws OrchestraException
	 *
	 * @throws OrchestraException
	 *             impossible to deploy the process
	 */
	public void deployBPELProcess() throws OrchestraException {
		if (this.bpelProcess == null) {
			this.logger.info("Open the BPEL context");
			this.openContext();
			this.logger.info("Deployment of BPEL process");
			this.bpelDeployer = new Deployer();
			// this.bpelDeployment = new Deployment();

			List<URL> wsdls = null;
			if (this.wsdlName != null) {
				wsdls = new ArrayList<URL>();
				wsdls.add(this.wsdlName);
			}
			Parse parse = new Parse();
			this.bpelDeployment = BpelImportParser.parseBpelFromUrl(this.bpelProcessName, wsdls, parse);

			// this.bpelDeployment.setWsdl(wsdlName);

			// this.bpelDeployment.setBpelUrl(bpelProcessName);
			this.bpelProcess = this.bpelDeployer.deploy(this.bpelDeployment, this.jbiEnvironment, parse);

			if (this.bpelProcess == null) {
				throw new OrchestraException("Impossible to deploy the process");
			}
		}
	}

	/**
	 * Undeploy the bpel process
	 *
	 * @throws MessagingException
	 *
	 * @throws OrchestraException
	 *             impossible to undeploy the process
	 */
	public void undeployBPELProcess() throws MessagingException, OrchestraException {
		try {
			if (this.bpelProcess != null) {
				this.bpelDeployer.undeploy(this.bpelProcess.getQName(), this.jbiEnvironment);
				this.logger.info("BPEL process Undeployed");
				this.closeContext();
				this.logger.info("BPEL context closed");
				this.bpelProcess = null;
			}
		} catch (final Exception e) {
			throw new MessagingException(e);
		}
	}

	/**
	 * Launch the bpel process
	 *
	 * @param exchange
	 *            the message exchange
	 * @param listener
	 *            the listener
	 * @throws MessagingException
	 *             impossible to start the process
	 * @throws WSDLException
	 *             impossible to read the WSDL
	 */
	public void launchBPELProcess(final Exchange exchange, final JBIListener listener)
	throws MessagingException, WSDLException {
		this.logger.info("Launch the BPEL process");

		if (this.bpelProcess == null) {
			throw new MessagingException("the Bpel process cannot be null");
		}
		if (exchange == null) {
			throw new MessagingException("no exchange");
		}

		// Initialization of the JBI invoker
		// TODO: I think that We must realize a clone of JBIInvoker
		final JBIInvoker serviceInvoker = (JBIInvoker) this.jbiEnvironment.get(Invoker.DEFAULT_KEY);
		serviceInvoker.init(listener);
		serviceInvoker.setSecuritySubjet(exchange.getInMessage().getSecuritySubject());

		this.logger.info("called endpoint service: " + exchange.getEndpoint().getServiceName());
		this.logger.info("called operation: " + exchange.getOperationName());

		if (exchange.getEndpoint().getInterfaces() == null
				|| exchange.getEndpoint().getInterfaces().length == 0) {
			throw new MessagingException("Cannot found the interface");
		}
		final QName processQName = new QName(exchange.getEndpoint().getServiceName()
				.getNamespaceURI(), exchange.getEndpoint().getServiceName().getLocalPart());
		final QName portTypeQName = new QName(exchange.getEndpoint().getInterfaces()[0]
		                                                                             .getNamespaceURI(), exchange.getEndpoint().getInterfaces()[0].getLocalPart());
		final String operationName = exchange.getOperationName();

		// Get the input message
		final PortType port = this.wsdlDef.getPortType(portTypeQName);

		if (port == null) {
			throw new MessagingException("Impossible to find the corresponding port: "
					+ portTypeQName);
		}

		final Operation operation = port.getOperation(operationName, null, null);

		MessageCarrier messageCarrier = null;

		final OperationKey operationKey = new OperationKey(processQName, portTypeQName,
				operationName);

		if (operation == null) {
			throw new MessagingException("This operation (" + operationName
					+ ") does not exist in WSDL document: " + this.wsdlName);
		}

		final QName inputMessageQName = new QName(operation.getInput().getMessage().getQName()
				.getNamespaceURI(), operation.getInput().getMessage().getQName().getLocalPart());
		this.logger.finest("processQName: " + processQName);
		this.logger.finest("portTypeQName: " + portTypeQName);
		final javax.wsdl.Message inputMessage = this.bpelProcess.getWsdlInfos().getMessage(
				inputMessageQName);
		this.logger.finest("input jbi message: " + inputMessage);

		// Convert the JBI message to BPEL message
		final Message incomingMessage = Marshaller.getInstance().buildBPELMessageFromJBIMessage(
				this.bpelProcess.getWsdlInfos(), portTypeQName, operation,
				inputMessage.getParts().values(), exchange.getInMessageContentAsDocument());
		this.logger.finest("converted bpel message: " + incomingMessage);

		try {
			final boolean requestResponse = this.isRequestResponse(this.bpelProcess, portTypeQName,
					operationName);

			this.logger.finest("is requestResponse: " + requestResponse);
			if (this.isRequestResponse(this.bpelProcess, portTypeQName, operationName)) {
				messageCarrier = new MessageCarrier();
			}
		} catch (final OrchestraException e) {
			if (this.jbiEnvironment != null) {
				final Transaction t = this.jbiEnvironment.getTransaction();
				if (t != null) {
					t.setRollbackOnly();
				}
			}
			throw new MessagingException(e);
		}


		try {
			this.logger.finest("handle process...");
			final BpelExecution bpelExecution = Receiver.handle(incomingMessage, operationKey,
					this.environmentFactory, messageCarrier);
			this.logger.finest("process finished...");
			BpelExecution instance = null;
			if (bpelExecution != null) {
				instance = bpelExecution.getProcessInstance();
			}
			this.logger.finest("instance of process: " + instance);

			// Get the response if exist
			final Message responseMessage = messageCarrier.getMessage();

			this.logger.finest("responseMessage: " + responseMessage);
			if (responseMessage != null) {
				this.logger.finest("bpel response message = " + responseMessage);

				// We make the response
				final Collection<Part> values = operation.getOutput().getMessage().getParts()
				.values();
				final Document jbiDocument = Marshaller.getInstance()
				.buildJBIMessageFromBPELMessage(this.bpelProcess.getWsdlInfos(),
						portTypeQName, responseMessage, operation, values,
						port.getQName().getNamespaceURI());

				if (exchange.getPattern().equals(MEPConstants.IN_OUT_PATTERN.value())
						|| exchange.getPattern().equals(
								MEPConstants.IN_OPTIONAL_OUT_PATTERN.value())) {
					exchange.setOutMessageContent(jbiDocument);
				} else {
					throw new MessagingException("Incorrect MEP pattern");
				}
			}
		} catch (final Exception e) {
			System.err.println("error: " + e.getMessage());
			e.printStackTrace();
			throw new MessagingException(e);
		}
	}

	/**
	 * Return true if the operation corresponding to the given parameters is a
	 * requestResponse operation.
	 *
	 * @param bpelProcess1
	 *            bpelProcess
	 * @param portTypeQName
	 *            portTypeName
	 * @param operationName
	 *            operationName
	 * @return true if the operation corresponding to the given parameters is a
	 *         requestResponse operation
	 * @throws OrchestraException
	 */
	@SuppressWarnings("unchecked")
	private boolean isRequestResponse(final BpelProcess bpelProcess1, final QName portTypeQName,
			final String operationName) throws OrchestraException {
		final PortType pt = bpelProcess1.getWsdlInfos().getPortType(portTypeQName);
		if (pt == null) {
			throw new OrchestraException("Process '" + bpelProcess1.getName() + "', portType '"
					+ portTypeQName + "' not found : cannot define if it is a request/response"
					+ " call or a oneWay call.");
		}
		final List<Operation> operations = pt.getOperations();
		for (final Operation op : operations) {
			if (op.getName().equals(operationName)) {
				return op.getOutput() != null;
			}
		}
		throw new OrchestraException("Operation '" + operationName + "' not found in portType '"
				+ portTypeQName
				+ "' : cannot define if it is a request/response call or a oneWay call.");
	}

	/**
	 * Open the JBI context
	 */
	private void openContext() {
		this.jbiEnvironment = this.environmentFactory.openEnvironment();
		this.bpelDeployment = new Deployment();

		final List<URL> wsdls = new ArrayList<URL>();

		for (final URL externalWsdl : this.externalsWsdls) {
			wsdls.add(externalWsdl);
		}
		if (this.wsdlName != null) {
			wsdls.add(this.wsdlName);
		}
		Parse parse = new Parse();
		this.bpelDeployment = BpelImportParser.parseBpelFromUrl(this.bpelProcessName, wsdls, parse);

		// Add wsdl and bpel process
		// this.bpelDeployment.setWsdl(this.wsdlName);
		// this.bpelDeployment.setBpelUrl(this.bpelProcessName);

		// Ajout des WSDL externes
		// for (final URL externalWsdl : this.externalsWsdls) {
		// this.bpelDeployment.addExternalWsdl(externalWsdl);
		// }
	}

	/**
	 * Close the context
	 */
	private void closeContext() {
		this.jbiEnvironment.close();
		this.bpelProcess = null;
	}

}
