/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra;

import java.io.File;
import java.io.FileFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;

import org.jbpm.env.EnvironmentFactory;
import org.ow2.petals.component.framework.Constants.Component;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.su.ServiceUnitListener;
import org.ow2.petals.jbi.descriptor.Extensions;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.descriptor.Provides;

/**
 * This class implements the orchestra su listener
 *
 * @author Nicolas SALATGE - eBMWebSourcing
 *
 */
public final class OrchestraSUListener implements ServiceUnitListener {

    /**
     * BPEL property
     */
    private static final String BPEL_PROPERTY = "bpel";

    /**
     * JDK logger
     */
    private final Logger logger;

    /**
     * JBI Descriptor
     */
    private JBIDescriptor descriptor;

    /**
     * The orchestra SU Manager
     */
    private final OrchestraSUManager suManager;

    /**
     * The environment Factory
     */
    private final EnvironmentFactory environmentFactory;

    /**
     * The su root path
     */
    private String suRootPath;

    /**
     * The default constructor
     *
     * @param environmentFactory
     *            the jbi environnement
     * @param suManager
     *            the su manager
     * @param logger
     *            the logger
     */
    public OrchestraSUListener(final OrchestraSUManager suManager,
            final EnvironmentFactory environmentFactory, final Logger logger) {
        this.logger = logger;
        this.suManager = suManager;
        this.environmentFactory = environmentFactory;
    }

    /**
     * This method is realize when the su is deployed
     *
     * @param serviceUnitName
     *            the name of the service unit
     * @param jbiSuRootPath
     *            the su root path
     * @param jbiDescriptor
     *            the JBI descriptor
     * @throws PEtALSCDKException
     *             Impossible to realize this method
     */
    public void onSUDeployed(final String serviceUnitName, final String jbiSuRootPath,
            final JBIDescriptor jbiDescriptor) throws PEtALSCDKException {

        this.logger.info("deployment of the SU " + serviceUnitName + "...");
        this.descriptor = jbiDescriptor;
        this.suRootPath = jbiSuRootPath;
        this.createBPELContextsForEachProvides(this.descriptor);

        // Deploy the BPEL process
        final List<Provides> providesList = jbiDescriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            final BPELContext element = this.suManager.get(provides.getEndpointName());
            element.deployBPELProcess();
            this.logger.info("Service endpoint deployed: " + provides.getEndpointName());
        }

        this.logger.info("SU deployed: " + serviceUnitName);
    }

    /**
     * Get external Wsdl Urls
     *
     * @param directoryName
     *            the name of the directory
     * @param mainWsdlName
     *            the main wsdl associated to the process
     * @return the list of externals wsdl
     * @throws MalformedURLException
     *             impossible to get external wsdls
     */
    public static ArrayList<URL> getExternalWsdlUrls(final String directoryName,
            final URL mainWsdlName) throws MalformedURLException {
        final File path = new File(directoryName);

        final FileFilter wsdlFileFilter = new FileFilter() {
            public boolean accept(final File f) {
                return f.getName().toLowerCase().endsWith(".wsdl");
            }
        };

        final ArrayList<URL> wsdls = new ArrayList<URL>();
        for (final File file : path.listFiles(wsdlFileFilter)) {
            if (!file.toURL().equals(mainWsdlName)) {
                wsdls.add(file.toURL());
            }
        }

        return wsdls;
    }

    /**
     * This method is realize when the su is started
     *
     * @param serviceUnitName
     *            the name of the service unit
     * @throws OrchestraException
     *             Impossible to realize this method
     */
    public void onSUStarted(final String serviceUnitName) throws PEtALSCDKException {
    }

    /**
     * this method is realize when the su is stopped
     *
     * @param serviceUnitName
     *            the name of the service unit
     * @throws OrchestraException
     *             Impossible to realize this method
     */
    public void onSUStopped(final String serviceUnitName) throws PEtALSCDKException {
    }

    /**
     * this method is realize when the su is undeployed
     *
     * @param serviceUnitName
     *            the name of the service unit
     * @param jbiSuRootPath
     *            the su root path
     * @param jbiDescriptor
     *            the JBI descriptor
     * @throws PEtALSCDKException
     *             Impossible to realize this method
     */
    public void onSUUndeployed(final String serviceUnitName, final String jbiSuRootPath,
            final JBIDescriptor jbiDescriptor) throws PEtALSCDKException {
        this.logger.info("undeployment of the SU " + serviceUnitName + "...");

        // Undeploy the BPEL process
        final List<Provides> providesList = jbiDescriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            final BPELContext element = this.suManager.remove(provides.getEndpointName());
            try {
                element.undeployBPELProcess();
            } catch (final MessagingException e) {
                throw new PEtALSCDKException(e);
            }
            this.logger.info("service endpoint deployed: " + provides.getEndpointName());
        }

        this.logger.info("SU undeployed: " + serviceUnitName);
    }

    /**
     * Create a BPEL context for each provides included in the SU
     *
     * @param jbiDescriptor
     *            the JBIDescriptor
     * @throws PEtALSCDKException
     *             impossible to generate urls
     * @throws OrchestraException
     */
    private void createBPELContextsForEachProvides(final JBIDescriptor jbiDescriptor)
        throws PEtALSCDKException {
        final List<Provides> providesList = jbiDescriptor.getServices().getProvides();
        try {
            BPELContext bpelContext = null;
            for (final Provides provides : providesList) {
                final org.ow2.petals.component.framework.util.Extensions cfExtensions
                    = new org.ow2.petals.component.framework.util.Extensions(
                        provides.getExtensions());
                final Properties props = cfExtensions.getPetalsExtensions();

                // get parameters from extension
                final String bpelProcessName = props.getProperty(OrchestraSUListener.BPEL_PROPERTY);
                final String wsdlName = props.getProperty(Component.WSDL_LOCATION);

                final ArrayList<URL> externalWsdls = OrchestraSUListener.getExternalWsdlUrls(
                        this.suRootPath, OrchestraSUListener.getResourceAsURL(this.suRootPath,
                                wsdlName));

                bpelContext = this.createBPELContext(bpelProcessName, wsdlName, provides
                        .getExtensions(), externalWsdls);

                // Store the process
                this.suManager.add(provides.getEndpointName(), bpelContext);
            }
        } catch (final MalformedURLException e) {
            throw new PEtALSCDKException(e.getMessage());
        }
    }

    /**
     * Create a BPEL context form the extensions includes in the provides
     * elements of the SU
     *
     * @param extensions
     *            the extensions
     * @param externalWsdls
     *         list of all external WSDLs used by the process
     * @throws MalformedURLException
     *             impossible to generate URLs
     * @return the BPEL context
     */
    private BPELContext createBPELContext(final String bpelProcessName, final String wsdlName,
            final Extensions extensions, final ArrayList<URL> externalWsdls)
        throws MalformedURLException {

        if (bpelProcessName == null || bpelProcessName.equals("")) {
            throw new IllegalArgumentException("the BPEL process cannot be null in jbi descriptor");
        }
        if (wsdlName == null || wsdlName.equals("")) {
            throw new IllegalArgumentException("the WSDL name cannot be null in jbi descriptor");
        }

        return new BPELContext(OrchestraSUListener.getResourceAsURL(this.suRootPath,
                bpelProcessName), OrchestraSUListener.getResourceAsURL(this.suRootPath, wsdlName),
                this.environmentFactory, this.logger, externalWsdls);
    }

    /**
     * We retrieve the URL associated to the resource
     *
     * @param resourceName
     *            the resource name
     * @return the URL
     * @throws MalformedURLException
     *             impossible to generate url
     */
    private static URL getResourceAsURL(final String rootPath, final String resourceName)
        throws MalformedURLException {
        return new File(rootPath + "/" + resourceName).toURL();
    }

}
