/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.engine.orchestra.jbi;

import java.util.Map;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.PortType;
import javax.wsdl.WSDLException;
import javax.xml.namespace.QName;

import org.jbpm.env.CurrentEnvironment;
import org.ow2.orchestra.execution.BpelProcess;
import org.ow2.orchestra.execution.runtime.PartnerLinkRuntime;
import org.ow2.orchestra.execution.runtime.VariableRuntime;
import org.ow2.orchestra.execution.services.OperationKey;
import org.ow2.orchestra.execution.services.itf.Invoker;
import org.ow2.orchestra.execution.services.itf.Repository;
import org.ow2.orchestra.var.Message;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.engine.orchestra.OrchestraException;
import org.ow2.petals.engine.orchestra.listener.JBIListener;
import org.ow2.petals.engine.orchestra.util.Marshaller;
import org.ow2.petals.engine.orchestra.util.Sender;

/**
 * This class implements the JBI Invoker for Orchestra
 *
 * @author Nicolas SALATGE - eBMWebSourcing
 */
public class JBIInvoker implements Invoker {

	/**
	 * The jbi component context
	 */
	private JBIListener listener;

	/**
	 * The security context
	 */
	private Subject securitySubject;

	/**
	 * Default constructor
	 */
	public JBIInvoker() {
		super();
		this.securitySubject = null;
	}

	/**
	 * Init the JBI INvoker
	 *
	 * @param e
	 *            the engine component
	 */
	public final void init(final JBIListener e) {
		this.listener = e;
	}

	/**
	 * Perform an invoke using the given parameters.
	 *
	 * @param operationKey
	 *            the key operation
	 * @param name
	 *            the name of the invoke activity
	 * @param partnerLinkRuntime
	 *            the partnerLink runtime value of the invoke
	 * @param variableRuntime
	 *            the value of the variable used in invoke
	 * @return the message received in response of this call (if there is
	 *         one...)
	 */
	public final Message invoke(final OperationKey operationKey, final String name,
			final PartnerLinkRuntime partnerLinkRuntime, final VariableRuntime variableRuntime) {
		Message messageReturned = null;

		if (this.listener == null) {
			throw new org.ow2.orchestra.exception.OrchestraException("The listener cannot be null");
		}

		try {
			this.listener.getLogger().log(Level.FINEST, "Entering in jbi invoke method");

			// We find the portType specified for the partner role
			final QName portTypeQName = operationKey.getPortTypeQName();
			final String operationName = operationKey.getOperationName();
			final QName processQName = operationKey.getProcessQName();

			// ProcessRepository processRepository =
			// CurrentEnvironment.get().get(Repository.class).getProcessRepository();
			final BpelProcess process = CurrentEnvironment.get().get(Repository.class).getProcess(
					processQName);

			final PortType portType = process.getWsdlInfos().getPortType(portTypeQName);

			// Get the Operation associated with the portType.
			final Operation operation = portType.getOperation(operationName, null, null);
			if (operation == null) {
				throw new org.ow2.orchestra.exception.OrchestraException("Operation "
						+ operationName + " not found");
			}

			this.listener.getLogger().log(Level.FINEST, "Operation name: " + operation.getName());

			this.listener.getLogger().log(Level.FINEST,
					"partnerLinkRuntime name: " + partnerLinkRuntime.getName());

			final Message message = (Message) variableRuntime.getValue();

			this.listener.getLogger().log(Level.FINEST,
					"Bpel message corresponding to the jbi request: " + message);

			Exchange request = null;

			request = Marshaller.getInstance().createJBIMessageExchangeFromBPELMessage(
					process.getWsdlInfos(), portTypeQName, message, operation,
					operation.getInput().getMessage().getParts().values(), portType.getQName(),
					this.listener, this.getSecuritySubjet());


			this.listener.getLogger().log(Level.FINEST, "Creation of Sender");
			final Sender sender = new Sender(this.listener, operation);

			final Exchange response = sender.send(partnerLinkRuntime.getPartnerRoleEndPointReference(), request, null);
			this.listener.getLogger().log(Level.INFO,
					"Receive response from: " + response.getEndpointName());
			if (operation.getOutput() != null) {
				final Map<String, Part> outputParts = operation.getOutput().getMessage().getParts();
				messageReturned = Marshaller.getInstance().buildBPELMessageFromJBIMessage(
						process.getWsdlInfos(), portTypeQName, operation, outputParts.values(),
						response.getOutMessageContentAsDocument());

				this.listener.getLogger().finest("bpel message returned: " + messageReturned);
			}

			this.listener.getLogger().log(Level.FINEST, "Exit of the jbi invoke method");
		} catch (final WSDLException e) {
			throw new org.ow2.orchestra.exception.OrchestraException("A WSDL exception was caught : ", e);
		} catch (final MessagingException e) {
			throw new org.ow2.orchestra.exception.OrchestraException(
					"Exception caught while invoke JBI Service", e);
		} catch (final OrchestraException e) {
			throw new org.ow2.orchestra.exception.OrchestraException(
					"Exception caught while invoke JBI Service", e);
		}
		return messageReturned;
	}

	/**
	 * Get the security subject
	 */
	public final Subject getSecuritySubjet() {
		return this.securitySubject;
	}

	/**
	 * Set the security subject
	 *
	 * @param securitySubject1
	 */
	public final void setSecuritySubjet(final Subject securitySubject1) {
		this.securitySubject = securitySubject1;
	}

}
