/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.engine.orchestra.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import javax.jbi.messaging.MessagingException;
import javax.security.auth.Subject;
import javax.wsdl.Operation;
import javax.wsdl.Part;
import javax.wsdl.WSDLException;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.var.Message;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.engine.orchestra.OrchestraException;
import org.ow2.petals.engine.orchestra.listener.JBIListener;
import org.ow2.petals.jbi.descriptor.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 *
 * @author nsalatge
 *
 */

public final class Marshaller {

	/**
	 * Mode
	 */
	public static enum ExchangeStyle {
		RPC, DOCUMENT
	};

	/**
	 * Singleton
	 */
	private static Marshaller marshaller;

	/**
	 * Default constructor
	 */
	private Marshaller() {

	}

	/**
	 * Get instance of the marshaller
	 *
	 * @return the marshaller
	 */
	public static Marshaller getInstance() {
		if (Marshaller.marshaller == null) {
			Marshaller.marshaller = new Marshaller();
		}
		return Marshaller.marshaller;
	}

	/**
	 * Create a JBI message exchange
	 *
	 * @param itf
	 *            the QName
	 * @param operation
	 *            the operation name
	 * @param content
	 *            the content of the operation
	 * @param attachmentFiles
	 *            the list of attachments
	 * @param listener
	 *            the listener
	 * @param properties
	 *            the properties
	 * @param subject
	 *            the orchestra subject
	 * @return the exchange. May be null
	 * @throws MessagingException
	 */
	private Exchange createMsg(final QName itf, final Operation operation, final Document content,
			final List<File> attachmentFiles, final String properties, final JBIListener listener,
			final Subject subject) throws MessagingException {

		MEPConstants mep = MEPConstants.IN_ONLY_PATTERN;

		if (operation.getOutput() != null) {
			mep = MEPConstants.IN_OUT_PATTERN;
		}

		final Exchange exchange = listener.createExchange(mep);
		exchange.setInterfaceName(itf);

		if (exchange != null) {
			exchange.setOperation(QName.valueOf(operation.getName()));
			// setting content
			exchange.setInMessageContent(content);

			if (subject != null) {
				exchange.getInMessage().setSecuritySubject(subject);

			}

			// TODO: Copy attachment and other properties
		}

		return exchange;
	}

	/**
	 * Build JBI message from BPEL message
	 *
	 * @param wsdl
	 *            the WSDL
	 * @param message
	 *            the BPEL Message
	 * @param portTypeQName
	 *            the portTypeQName
	 * @param operation
	 *            the operation
	 * @param wsdlPart
	 *            the orchestra WSDL collection
	 * @param itf
	 *            the QName
	 * @param listener
	 *            the orchestra listener
	 * @param subject
	 *            the orchestra subject
	 * @return the jbi message exchange
	 * @throws OrchestraException
	 *             impossible to build JBI message
	 * @throws WSDLException
	 */
	public Exchange createJBIMessageExchangeFromBPELMessage(final WsdlsInfos wsdl,
			final QName portTypeQName, final Message message, final Operation operation,
			final Collection<Part> wsdlPart, final QName itf, final JBIListener listener,
			final Subject subject) throws OrchestraException, WSDLException {
		Exchange exchange = null;
		try {
			final Document document = this.buildJBIMessageFromBPELMessage(wsdl, portTypeQName,
					message, operation, wsdlPart, itf.getNamespaceURI());

			try {
				listener.getLogger().info(
						"Converted JBI message: "
						+ XMLUtil.parseToString(this.buildJBIMessageFromBPELMessage(wsdl,
								portTypeQName, message, operation, wsdlPart, itf
								.getNamespaceURI())));
			} catch (final TransformerException e) {
				listener.getLogger().info("Error when print the converted jbi message");
			}
			exchange = this.createMsg(itf, operation, document, null, null, listener, subject);

		} catch (final MessagingException e) {
			throw new OrchestraException("Exception while building a JBI Message", e);
		}
		return exchange;
	}

	/**
	 * Build JBI Document from BPEL message
	 *
	 * @param wsdl
	 *            the WSDL
	 * @param portTypeQName
	 *            the portTypeQName
	 * @param message
	 *            the BPEL Message
	 * @param operation
	 *            the operation
	 * @param wsdlPart
	 *            the WSDL collection
	 * @param operationNamespace
	 *            the operationNamespace
	 * @return the JBI document
	 * @throws OrchestraException
	 *             impossible to build JBI document
	 * @throws WSDLException
	 */
	public Document buildJBIMessageFromBPELMessage(final WsdlsInfos wsdl,
			final QName portTypeQName, final Message message, final Operation operation,
			final Collection<Part> wsdlPart, final String operationNamespace)
	throws OrchestraException, WSDLException {
		Document document = null;
		try {

			WSDLAnalyzer analyzer;

			analyzer = new WSDLAnalyzer(wsdl);

			final ExchangeStyle style = analyzer.getStyleForOperation(portTypeQName, operation
					.getName());

			if (style == null) {
				throw new OrchestraException(
				"Impossible to find the style of exchange (RPC or Document)");
			}

			document = BpelUtil.getNewDocument();

			final List<Element> parts = this.getListOfParts(message, document);

			if (style == ExchangeStyle.RPC) {
				final Element operationElmt = document.createElementNS(operationNamespace,
						operation.getName());
				operationElmt.setPrefix("s");
				for (final Element part : parts) {
					Element elmt = null;
					elmt = this.importElementInJbiMessage(portTypeQName, wsdlPart, document, part,
							style);
					if (elmt == null) {
						throw new OrchestraException("Impossible to find this part message "
								+ part.getNodeName() + " in operation " + operation.getName()
								+ " in WSDL");
					}
					operationElmt.appendChild(elmt);
				}
				document.appendChild(operationElmt);
			} else {
				final List<Element> elmts = new ArrayList<Element>();
				for (final Element part : parts) {
					Element elmt = null;
					elmt = this.importElementInJbiMessage(portTypeQName, wsdlPart, document, part,
							style);
					if (elmt == null) {
						throw new OrchestraException("Impossible to find this part message "
								+ part.getNodeName() + " in operation " + operation.getName()
								+ " in wsdl");
					}

					elmts.add(elmt);

				}

				if (elmts.size() > 1) {
					throw new OrchestraException("Impossible to have several root elements");
				}

				if (elmts.size() == 1) {
					document.appendChild(elmts.get(0));
				}
			}
		} catch (final OrchestraException e) {
			throw new OrchestraException("Exception while building JBI Message", e);
		}
		return document;
	}

	private Element importElementInJbiMessage(final QName portTypeQName,
			final Collection<Part> wsdlPart, final Document document, final Element part,
			final ExchangeStyle style) {
		Element elmt = null;
		for (final Part p : wsdlPart) {
			if (p.getName().equals(part.getNodeName())) {
				if (p.getElementName() != null) {
					elmt = document.createElementNS(p.getElementName().getNamespaceURI(), p
							.getElementName().getLocalPart());
				} else {
					if (style == ExchangeStyle.RPC) {
						elmt = document.createElement(p.getName());
					} else {
						elmt = document.createElementNS(portTypeQName.getNamespaceURI(), p
								.getName());
					}
				}
				if (part.getFirstChild() != null
						&& part.getFirstChild().getNodeType() == Node.TEXT_NODE
						&& part.getChildNodes().getLength() == 1) {
					elmt.setTextContent(part.getTextContent());

				} else {
					for (int i = part.getChildNodes().getLength() - 1; i >= 0; i--) {
						final Node newNode = elmt.getOwnerDocument().adoptNode(
								part.getChildNodes().item(i).cloneNode(true));
						elmt.appendChild(newNode);
					}
				}

				break;
			}
		}
		return elmt;
	}

	private List<Element> getListOfParts(final Message message, final Document document) {
		final List<Element> parts = new ArrayList<Element>();
		final Iterator<Entry<String, Element>> it = message.getParts().entrySet().iterator();
		while (it.hasNext()) {
			final Entry<String, Element> entry = it.next();
			final String partName = entry.getKey();
			final Element partElement = entry.getValue();

			final Element partEl = document.createElement(partName);

			if (partElement.getFirstChild() != null
					&& partElement.getFirstChild().getNodeType() == Node.TEXT_NODE
					&& partElement.getChildNodes().getLength() == 1) {
				partEl.setTextContent(partElement.getTextContent());

			} else {
				for (int i = partElement.getChildNodes().getLength() - 1; i >= 0; i--) {
					final Node newNode = partEl.getOwnerDocument().adoptNode(
							partElement.getChildNodes().item(i).cloneNode(true));
					partEl.appendChild(newNode);
				}

			}
			parts.add(partEl);
		}
		return parts;
	}

	/**
	 * Build BPEL message from JBI message exchange
	 *
	 * @param parts
	 *            the bpel parts element
	 * @param jbiMessage
	 *            the jbi message
	 * @return the BPEL message
	 * @throws MessagingException
	 *             impossible to build BPEL Message
	 */
	public Message buildBPELMessageFromJBIMessage(final WsdlsInfos wsdl, final QName portTypeQName,
			final Operation operation, final Collection<Part> parts, final Document jbiMessage)
	throws MessagingException {
		Message message = null;
		try {
			final HashMap<String, Element> messageParts = new HashMap<String, Element>();

			for (final Part part : parts) {

				Element el = null;

				WSDLAnalyzer analyzer;

				analyzer = new WSDLAnalyzer(wsdl);

				final ExchangeStyle style = analyzer.getStyleForOperation(portTypeQName, operation
						.getName());

				if (style == null) {
					throw new OrchestraException(
					"Impossible to find the style of exchange (rpc or document)");
				}

				if (part.getElementName() != null) {
					el = this.getPartOrElement(jbiMessage.getDocumentElement(), part
							.getElementName().getLocalPart(), style);
				} else {
					el = this.getPartOrElement(jbiMessage.getDocumentElement(), part.getName(), style);
				}


				if (el == null) {

					String partOrElmt = null;
					if(part.getElementName() != null) {
						part.getElementName().getLocalPart();
					}
					if(partOrElmt == null) {
						partOrElmt = part.getName();
					}
					throw new MessagingException("the part or element named "
							+ partOrElmt + " is not found in the request");
				}

				final Element newElement = this.copyElementWithNs(part, el);

				if (part.getTypeName() != null
						&& BpelUtil.isXmlSchemaNs(part.getTypeName().getNamespaceURI())) {
					// it is a schema type
					String text = null;
					final Node child = newElement.getFirstChild();

					if (child != null) {
						if(child.getNodeType() == Node.TEXT_NODE) {
							text = ((Text)child).getData();
						} else if(child.getNodeType() == Node.ELEMENT_NODE) {
							text = ((Element)child).getTextContent();
						}
					} else {
						text = newElement.getTextContent();
					}
					final Element element = BpelXmlUtil.createElementWithContent(text);
					messageParts.put(part.getName(), element);
				} else {
					messageParts.put(part.getName(), newElement);
				}

			}

			message = new Message(messageParts);

		} catch (final WSDLException e) {
			throw new MessagingException("Impossible to convert BPEL message from JBI message", e);
		} catch (final OrchestraException e) {
			throw new MessagingException("Impossible to convert BPEL message from JBI message", e);
		}
		return message;
	}

	/**
	 * Copy of elements with their name space
	 *
	 * @param part
	 * @param el
	 * @return
	 * @throws MessagingException
	 */
	private Element copyElementWithNs(final Part part, final Element el) throws MessagingException {
		Element newElement = null;

		if (part.getTypeName() != null) {
			newElement = BpelXmlUtil.createElement(part.getTypeName().getNamespaceURI(), part
					.getName());
		} else if (part.getElementName() != null) {
			newElement = BpelXmlUtil.createElement(part.getElementName().getNamespaceURI(), part
					.getName());
		}

		if (newElement == null) {
			throw new MessagingException(
					"Impossible to find the type or the element associate to this part: "
					+ part.getName());
		}

		for (int i = 0; i < el.getChildNodes().getLength(); i++) {
			final Node n = newElement.getOwnerDocument().adoptNode(
					el.getChildNodes().item(i).cloneNode(true));
			newElement.appendChild(n);
		}

		return newElement;
	}

	/**
	 * Find the part or the element of a father element from the father element
	 *
	 * @param father
	 *            the father element
	 * @param partOrElmtName
	 *            the part / element name
	 * @return return null
	 */
	// TODO correct the number of 'return' (put only one instead of 3)
	private Element getPartOrElement(final Element father, final String partOrElmtName, final ExchangeStyle style) {

		if ((style != ExchangeStyle.RPC)&&(father.getLocalName().equals(partOrElmtName))) {
			return father;
		}

		for (int i = 0; i < father.getChildNodes().getLength(); i++) {
			if (father.getChildNodes().item(i).getLocalName() != null
					&& father.getChildNodes().item(i).getLocalName().equals(partOrElmtName)) {
				return (Element) father.getChildNodes().item(i);
			}
		}
		return null;
	}

}
