/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.engine.orchestra.util;

import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.wsdl.Operation;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.engine.orchestra.listener.JBIListener;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;

/**
 *
 * @author nsalatge
 *
 */

public class Sender {

	/**
	 * Send style
	 *
	 */
	public static enum SendStyle {
		ASYNCHRONOUS_SEND, SYNCHRONOUS_SEND, SYNCHRONOUS_SEND_WITH_TIMEOUT
	};

	/**
	 * Style
	 */
	private SendStyle sendStyle;

	/**
	 * The jbi listener
	 */
	private final JBIListener listener;

	/**
	 * Default constructor
	 *
	 * @throws MessagingException
	 */
	public Sender(final JBIListener listener,
			final Operation operation) throws MessagingException {
		this.sendStyle = SendStyle.ASYNCHRONOUS_SEND;
		if (operation.getOutput() != null) {
			this.sendStyle = SendStyle.SYNCHRONOUS_SEND;
		}
		this.listener = listener;

	}

	/**
	 * Call a jbi service
	 *
	 * @param partner
	 *            the partner to invoke. May be null
	 * @param msgIn
	 *            the exchange
	 * @param syncTime
	 *            the time if it a send synchrone type
	 * @return the exchange
	 * @throws MessagingException
	 *             impossible to send exchange
	 */
	public final Exchange send(final Element partner, final Exchange msgIn,
			final Long syncTime) throws MessagingException {
		this.listener.getLogger().finest("Entering in send method");
		if(partner != null) {
			String endpointName = this.getEndPointName(partner);
			QName service = this.getService(partner);

			if(service != null) {
				this.listener.getLogger().info("Set the service to invoke: " + service);
				msgIn.setService(service);
				if(endpointName != null) {
					ServiceEndpoint endpoint = this.resolveEndPoint(service, endpointName);
					this.listener.getLogger().info("Set the endpoint to invoke: " + endpoint.getEndpointName());
					msgIn.setEndpoint(endpoint);
				}
			}
		} else {
			this.listener.getLogger().finest("partner is null: research the service by its interface");
		}

		if (this.sendStyle == SendStyle.ASYNCHRONOUS_SEND) {
			this.listener.send(msgIn);
		} else if (this.sendStyle == SendStyle.SYNCHRONOUS_SEND) {
			this.listener.sendSync(msgIn);
		} else if (this.sendStyle == SendStyle.SYNCHRONOUS_SEND_WITH_TIMEOUT) {
			this.listener.sendSync(msgIn, syncTime);
		} else {
			throw new MessagingException("this type of message ("
					+ this.sendStyle + ") is unknown");
		}

		if (msgIn.isErrorStatus()) {
			throw new MessagingException("Error from the provider "
					+ msgIn.getEndpointName() + ": " + msgIn.getError());
		}
		this.listener.getLogger().finest("Exit to the send method");
		return msgIn;
	}

	/**
	 * Extract the service from a partner reference information item.
	 *
	 * @param ii
	 *            Information Item
	 * @return the service QName
	 * @throws MessagingException
	 *             problem to consult the annuary
	 */
	public final QName getService(final Element ii)
	throws MessagingException {

		listener.getLogger().log(Level.FINEST,
		"Entering in getService method");

		final org.w3c.dom.Element serviceRefElem = (org.w3c.dom.Element) ii;

		final org.w3c.dom.Element endpointRef = (org.w3c.dom.Element) serviceRefElem
		.getElementsByTagName("EndpointReference").item(0);

		final org.w3c.dom.Element serviceElmt = (org.w3c.dom.Element) endpointRef
		.getElementsByTagName("ServiceName").item(0);

		String serviceName = null;
		String prefixName = null;

		if (serviceElmt.getTextContent().indexOf(":") > -1) {
			serviceName = serviceElmt.getTextContent().substring(
					serviceElmt.getTextContent().indexOf(":") + 1);
			prefixName = serviceElmt.getTextContent().substring(0,
					serviceElmt.getTextContent().indexOf(":"));
		} else {
			serviceName = serviceElmt.getTextContent();
		}

		String serviceNamespace = serviceElmt.lookupNamespaceURI(prefixName);
		if (serviceNamespace == null) {
			String prefixTmp = null;
			final NamedNodeMap list = serviceElmt.getAttributes();
			for (int i = 0; i < list.getLength(); i++) {

				if (list.item(i).getNodeName().indexOf(":") > -1) {
					prefixTmp = list.item(i).getNodeName().substring(
							list.item(i).getNodeName().indexOf(":") + 1);
				} else {
					prefixTmp = list.item(i).getNodeName();
				}

				if ((prefixName != null) && (prefixName.equals(prefixTmp))) {
					serviceNamespace = list.item(i).getNodeValue();
				}
			}
		}

		// TODO: Normally, just this line is sufficient
		// QName service = new QName(serviceElmt.lookupNamespaceURI(prefixName),
		// serviceName);

		final QName service = new QName(serviceNamespace, serviceName);

		listener.getLogger().log(Level.FINEST, "service name: " + service);
		listener.getLogger().log(Level.FINEST,
		"Exit of the getService method");
		return service;
	}

	/**
	 * Extract the endPoint reference from a service reference information item.
	 *
	 * @param ii
	 *            Information Item
	 * @return the end Point URL
	 * @throws MessagingException
	 *             problem to consult the annuary
	 */
	public final String getEndPointName(final Element ii)
	throws MessagingException {
		String endpointName = null;
		listener.getLogger().log(Level.FINEST,
		"Entering in getEndPointName method");

		final org.w3c.dom.Element serviceRefElem = (org.w3c.dom.Element) ii;

		org.w3c.dom.Element endpointRef = null;

		if(serviceRefElem.getElementsByTagName("EndpointReference").getLength() > 0) {
			endpointRef = (org.w3c.dom.Element) serviceRefElem.getElementsByTagName("EndpointReference").item(0);
		}

		if(endpointRef != null) {
			org.w3c.dom.Element endpoint = null;
			if(endpointRef.getElementsByTagName("Address").getLength() > 0) {
				endpoint = (org.w3c.dom.Element) endpointRef.getElementsByTagName("Address").item(0);
			}


			if(endpoint != null) {
				endpointName = endpoint.getTextContent();
			}
		}
		listener.getLogger().log(Level.FINEST,
				"endpoint name found: " + endpointName);
		listener.getLogger().log(Level.FINEST,
		"Exit of the getEndPointName method");

		return endpointName;
	}




	/**
	 * Extract the endPoint reference from a service reference information item.
	 *
	 * @param ii
	 *            Information Item
	 * @return the end Point URL
	 * @throws MessagingException
	 *             problem to consult the annuary
	 */
	public final ServiceEndpoint resolveEndPoint(final QName service, final String endpointName)
	throws MessagingException {

		listener.getLogger().log(Level.FINEST,
		"Entering in resolveEndPoint method");

		listener.getLogger().log(Level.FINEST, "service name: " + service);
		listener.getLogger().log(Level.FINEST, "endpoint name: " + endpointName);


		// Search by endpoint
		ServiceEndpoint endpoint = listener.getComponent().getContext().getEndpoint(service, endpointName);

		if (endpoint == null) {
			throw new MessagingException("endpoint (" + endpointName + ") for the service (" + service + ") not found.");
		}

		listener.getLogger().log(Level.FINEST,
				"endpoint service found: " + endpoint);
		listener.getLogger().log(Level.FINEST,
		"Exit of the resolveEndPoint method");

		return endpoint;
	}

}
