/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM WebSourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.engine.orchestra.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.wsdl.Binding;
import javax.wsdl.BindingOperation;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.ExtensibilityElement;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.xml.namespace.QName;

import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.petals.engine.orchestra.OrchestraException;
import org.ow2.petals.engine.orchestra.util.Marshaller.ExchangeStyle;

/**
 * this class allows to get critical information from WSDL definition like style
 * (rpc or document), mep to achieve a JBI call
 *
 * @author nsalatge
 *
 */
public class WSDLAnalyzer {

	/** The style parameter RPC */
	public static final String PARAM_RPC = "rpc";

	/** The style parameter Document */
	public static final String PARAM_DOCUMENT = "document";

	//TODO: use soap 1.1 and 1.2 namespace
	/** The name space URI used for the following three QName */
	public static final String NAMESPACE_URI = "http://schemas.xmlsoap.org/wsdl/soap/";

	/** The QName of the soap binding element in WSDL. */
	private static final QName SOAP_BINDING = new QName(WSDLAnalyzer.NAMESPACE_URI, "binding");

	/** The QName of the soap operation element in WSDL. */
	private static final QName SOAP_OPERATION = new QName(WSDLAnalyzer.NAMESPACE_URI, "operation");

	/** The QName of the soap body element in WSDL. */
	private static final QName SOAP_BODY = new QName(WSDLAnalyzer.NAMESPACE_URI, "body");

	/**
	 * The wsdl definition
	 */
	private final WsdlsInfos wsdl;

	/**
	 * Default constructor
	 *
	 * @throws WSDLException
	 */
	public WSDLAnalyzer(final WsdlsInfos wsdl) throws WSDLException {
		this.wsdl = wsdl;
	}

	/**
	 * Find services using the given portType (the interface)
	 *
	 * @param portTypeQName
	 *            the interface QName
	 * @return the list of services
	 * @throws OrchestraException
	 */
	public final List<Service> findServices(final QName portTypeQName) throws OrchestraException {
		return this.wsdl.getServicesOfPortType(portTypeQName);
	}

	/**
	 * Find the first service using the given portType (the interface) or null
	 *
	 * @param portTypeQName
	 *            the interface QName
	 * @return the list of services
	 * @throws OrchestraException
	 */
	public final Service findFirstCorrespondingService(final QName portTypeQName)
	throws OrchestraException {
		Service service = null;
		final List<Service> servicesList = this.findServices(portTypeQName);
		if (servicesList != null && servicesList.size() > 0) {
			service = servicesList.get(0);
		}
		return service;
	}

	/**
	 * Find bindings used in given service and using the given portType (the
	 * interface)
	 *
	 * @param service
	 *            the given service
	 * @param portTypeQName
	 *            the interface QName
	 * @return the list of services
	 * @throws OrchestraException
	 */
	public final List<Binding> getBindings(final Service service, final QName portTypeQName) {
		final List<Binding> bindings = new ArrayList<Binding>();
		if(service != null) {
			for (final Port port : (Collection<Port>) service.getPorts().values()) {
				final Binding binding = port.getBinding();
				final PortType portType = binding.getPortType();
				if (portType.getQName().equals(portTypeQName)) {
					bindings.add(binding);
				}
			}
		}
		return bindings;
	}

	/**
	 * Find the soap bindings used in given service and using the given portType
	 * (the interface)
	 *
	 * @param service
	 *            the given service
	 * @param portTypeQName
	 *            the interface QName
	 * @return the list of services
	 * @throws OrchestraException
	 */
	public final List<Binding> getSOAPBindings(final Service service, final QName portTypeQName) {
		final List<Binding> bindings = new ArrayList<Binding>();
		for (final Binding binding : this.getBindings(service, portTypeQName)) {
			// we look for the soap binding.
			for (final Object elem : binding.getExtensibilityElements()) {
				final QName elementQName = ((ExtensibilityElement) elem).getElementType();
				if (elementQName.equals(WSDLAnalyzer.SOAP_BINDING)) {
					bindings.add(binding);
				}
			}
		}
		return bindings;
	}

	/**
	 * Find the first soap bindings used in given service and using the given
	 * portType (the interface)
	 *
	 * @param service
	 *            the given service
	 * @param portTypeQName
	 *            the interface QName
	 * @return the list of services
	 * @throws OrchestraException
	 */
	public final Binding getFirstSOAPBindings(final Service service, final QName portTypeQName) {
		Binding binding = null;
		final List<Binding> bindingsList = this.getSOAPBindings(service, portTypeQName);
		if (bindingsList != null && bindingsList.size() > 0) {
			binding = bindingsList.get(0);
		}
		return binding;
	}

	/**
	 * Get the first binding corresponding to the port type
	 * @param portTypeQName
	 * @return the binding
	 * @throws OrchestraException
	 */
	private Binding getFirstBinding(final QName portTypeQName)
	throws OrchestraException {
		// We find the bindings for this portType
		final Service service = this.findFirstCorrespondingService(portTypeQName);

		Binding binding = null;
		if(service != null) {
			// We take the first binding of the WSDL definition which binds the
			// portType
			binding = this.getFirstSOAPBindings(service, portTypeQName);
		} else {
			List<Binding> bindings = this.wsdl.getBindingsOfPortType(portTypeQName);
			if(bindings.size() > 0) {
				// We take the first binding of the WSDL definition which binds the
				// portType
				binding = bindings.get(0);
			}
		}
		return binding;
	}

	/**
	 * Get the use (litteral or encoded) of operation
	 * @param portTypeQName
	 * @param operationName
	 * @return
	 * @throws OrchestraException
	 */
	public final String getUseForOperation(final QName portTypeQName, final String operationName)
	throws OrchestraException {
		// Use (RPC or document) used for the input of the operation
		String use = null;


		Binding binding = getFirstBinding(portTypeQName);


		// get particular style configuration for operation and URI of
		// soapAction
		final BindingOperation opBinding = binding.getBindingOperation(operationName, null,
				null);
		// get the use of the Input (Literal or encoded)
		for (final Object elem : opBinding.getBindingInput().getExtensibilityElements()) {
			final QName elementQName = ((ExtensibilityElement) elem).getElementType();
			if (elementQName.equals(WSDLAnalyzer.SOAP_BODY)) {
				final SOAPBody soapBody = (SOAPBody) elem;
				use = soapBody.getUse();
				break;
			}
		}

		return use;
	}

	/**
	 * Get the style (RPC or Document) of operation following the given
	 * interface and operation name
	 *
	 * @param portTypeQName
	 *            the interface
	 * @param operationName
	 *            the operation name
	 * @return the style or null if not found
	 * @throws OrchestraException
	 */
	public final ExchangeStyle getStyleForOperation(final QName portTypeQName, final String operationName)
	throws OrchestraException {
		String styleS = null;
		ExchangeStyle style = null;

		Binding binding = getFirstBinding(portTypeQName);

		// get particular style configuration for operation and URI of
		// soapAction
		final BindingOperation opBinding = binding.getBindingOperation(operationName, null, null);
		if(opBinding != null) {
			for (final Object elem : opBinding.getExtensibilityElements()) {
				final QName elementQName = ((ExtensibilityElement) elem).getElementType();
				if (elementQName.equals(WSDLAnalyzer.SOAP_OPERATION)) {
					final SOAPOperation soapOperation = (SOAPOperation) elem;
					if (soapOperation.getStyle() != null) {
						styleS = soapOperation.getStyle();
						if (styleS.equals(WSDLAnalyzer.PARAM_RPC)) {
							style = ExchangeStyle.RPC;
						} else if (styleS.equals(WSDLAnalyzer.PARAM_DOCUMENT)) {
							style = ExchangeStyle.DOCUMENT;
						}
					}
					break;
				}
			}
		}
		if (style == null) {
			styleS = this.getDefaultStyle(portTypeQName);
			if (styleS.equals(WSDLAnalyzer.PARAM_RPC)) {
				style = ExchangeStyle.RPC;
			} else if (styleS.equals(WSDLAnalyzer.PARAM_DOCUMENT)) {
				style = ExchangeStyle.DOCUMENT;
			}
		}

		return style;
	}


	/**
	 * Get the default style (Rpc or Document) of the operation
	 * @param portTypeQName
	 * @return
	 * @throws OrchestraException
	 */
	private String getDefaultStyle(final QName portTypeQName) throws OrchestraException {
		String defaultStyle = null;

		Binding binding = getFirstBinding(portTypeQName);

		// We look for the default style (Document) or (RPC) to use for
		// operations.
		for (final Object elem : binding.getExtensibilityElements()) {
			final QName elementQName = ((ExtensibilityElement) elem).getElementType();
			if (elementQName.equals(WSDLAnalyzer.SOAP_BINDING)) {
				final SOAPBinding soapBinding = (SOAPBinding) elem;
				defaultStyle = soapBinding.getStyle();
				break;
			}
		}

		return defaultStyle;
	}

}
